import six
from giotto import get_config

def make_tables():
    """
    Create all the tables for the models that have been added to the manifest.
    """
    engine = get_config('engine')
    Base = get_config('Base')
    Base.metadata.create_all(engine)
    print('Creating tables...')
    for i, t in enumerate(Base.metadata.tables.keys()):
        print(t)
    return '%s tables created' % (i + 1)

def blast_tables():
    """
    Drop all existing tables in the database, and then recreate them.
    """
    msg = "This will delete all data in your tables, are you sure? [yN]"
    if six.PY3:
        yn = input(msg)
    else:
        yn = raw_input(msg)
    if yn.lower() != 'y':
        return "Aborting"
    get_config('Base').metadata.drop_all(get_config('engine'))
    print("blasting away all tables...")
    return make_tables()


import inspect
import json

from giotto import get_config
from giotto.programs import GiottoProgram
from giotto.exceptions import (GiottoException, InvalidInput, ProgramNotFound,
    MockNotFound, ControlMiddlewareInterrupt, NotAuthorized)
from giotto.primitives import GiottoPrimitive
from giotto.keyvalue import DummyKeyValue
from giotto.control import GiottoControl

class GiottoController(object):
    middleware_interrupt = None
    persist_data = None

    def __init__(self, request, manifest, model_mock=False, errors=None):
        self.request = request
        self.model_mock = model_mock
        self.cache = get_config('cache', DummyKeyValue())
        self.errors = errors
        self.manifest = manifest
        self.middleware_interrupt_exc = None
        self.middleware_control = None
        self.display_data = 'Not calculated yet'
        
        # the program that corresponds to this invocation
        invocation = self.get_invocation()
        name = self.get_controller_name()
        parsed = self.manifest.parse_invocation(invocation, name)

        self.program = parsed['program']
        self.program.name = parsed['name']
        self.path_args = parsed['args']
        self.mimetype = parsed['superformat_mime'] or self.mimetype_override() or self.default_mimetype

    def get_response(self):
        try:
            last_good_request, middleware_result = self.program.execute_input_middleware_stream(self.request, self)
        except GiottoException as exc:
            # save this exception so it can be re-raised from within
            # get_data_response() so that get_concrete_response() can handle it
            self.middleware_interrupt_exc = exc
            self.request = last_good_request
        else:
            self.request = middleware_result # middleware ended cleanly

        if GiottoControl in type(middleware_result).mro():
            # middleware returned a control object
            self.middleware_control = middleware_result
            self.request = last_good_request

        response = self.get_concrete_response()

        if self.persist_data:
            response = self.persist(self.persist_data, response)

        return self.program.execute_output_middleware_stream(self.request, response, self)

    def get_data_response(self):
        """
        Execute the model and view, and handle the cache.
        Returns controller-agnostic response data.
        """
        if self.middleware_interrupt_exc:
            ## the middleware raised an exception, re-raise it here so
            ## get_concrete_response (defined in subclasses) can catch it.
            raise self.middleware_interrupt_exc

        if self.middleware_control:
            ## this redirect object came from middleware but return it as if it
            ## came from a view.
            return {'body': self.middleware_control}

        if self.model_mock and self.program.has_mock_defined():
            model_data = self.program.get_model_mock()
        else:
            args, kwargs = self.program.get_model_args_kwargs()
            data = self.get_data_for_model(args, kwargs)
            self.display_data = data # just for displaying in __repr__

            if self.program.cache and not self.errors:
                key = self.get_cache_key(data)
                hit = self.cache.get(key)
                if hit:
                    return hit
        
            model_data = self.program.execute_model(data)
        
        response = self.program.execute_view(model_data, self.mimetype, self.errors)

        if self.program.cache and not self.errors and not self.model_mock:
            self.cache.set(key, response, self.program.cache)

        if 'persist' in response:
            self.persist_data = response['persist']

        return response

    def get_data_for_model(self, args, kwargs):
        """
        In comes args and kwargs expected for the model. Out comes the data from
        this invocation that will go to the model.
        In other words, this function does the 'data negotiation' between the
        controller and the model.
        """
        raw_data = self.get_raw_data()
        defaults = kwargs
        values = args + list(kwargs.keys())

        output = {}
        i = -1
        for i, value in enumerate(values):
            if value in defaults:
                may_be_primitive = defaults[value]
                if isinstance(may_be_primitive, GiottoPrimitive):
                    output[value] = self.get_primitive(may_be_primitive.name)
                    continue # don't let user input override primitives.
                else:
                    output[value] = may_be_primitive

            if i + 1 <= len(self.path_args):
                output[value] = self.path_args[i]
            if value in raw_data:
                output[value] = raw_data[value]

        if len(self.path_args) > i + 1:
            raise ProgramNotFound("Too many positional arguments for program: '%s'" % self.program.name)

        if not len(output) == len(values):
            raise ProgramNotFound("Not enough data for program '%s'" % self.program.name)

        return output

    def persist(self, values):
        """
        Persist this data between the user and the server.
        """
        raise NotImplementedError

    def __repr__(self):
        controller = self.get_controller_name()
        model = self.program.name
        data = self.display_data
        return "<%s %s - %s - %s>" % (  
            self.__class__.__name__, controller, model, data
        )

    def mimetype_override(self):
        """
        In some circumstances, the returned mimetype can be changed. Return that here.
        Otherwise the default or superformat will be used.
        """
        return None

    def get_cache_key(self, data):
        try:
            controller_args = json.dumps(data, separators=(',', ':'), sort_keys=True)
        except TypeError:
            # controller contains info that can't be json serialized:
            controller_args = str(data)

        program = self.program.name
        return "%s(%s)(%s)" % (controller_args, program, self.mimetype)



from giotto.programs import GiottoProgram, ProgramManifest
from giotto.programs.shell import shell
from giotto.programs.tables import make_tables, blast_tables
from giotto.views import BasicView

management_manifest = ProgramManifest({
    'make_tables': GiottoProgram(
        controllers=['cmd'],
        model=[make_tables],
        view=BasicView
    ),
    'blast_tables': GiottoProgram(
        controllers=['cmd'],
        model=[blast_tables],
        view=BasicView,
    ),
    'shell': GiottoProgram(
        controllers=['cmd'],
        model=[shell],
        view=BasicView,
    ),
})


from __future__ import print_function
import json
import getpass

from .models import User
from giotto.exceptions import NotAuthorized
from giotto.control import Redirection
from giotto import get_config
from giotto.middleware import GiottoOutputMiddleware, GiottoInputMiddleware

class AuthenticationMiddleware(GiottoInputMiddleware):
    """
    This input middleware class must preceed any other authentication middleware class.
    It is used to extract authentiction information from the request, and
    verify that the credientials are correct.
    """
    def http(self, request):
        user = None
        session_key = request.cookies.get('giotto_session', None)
        if not session_key and request.POST:
            session_key = request.POST.get('auth_session', None)
        if session_key:
            username = get_config('auth_session').get(session_key)
            user = get_config('session').query(User).filter_by(username=username).first()

        setattr(request, 'user', user)
        return request

    def cmd(self, request):
        user = None
        session_key = request.enviornment.get('GIOTTO_SESSION', None)
        if session_key:
            user = get_config('auth_session').get(session_key)

        if not user:
            print("Username:")
            username = raw_input()
            password = getpass.getpass()
            user = User.get_user_by_password(username, password)

        setattr(request, 'user', user)
        return request

class AuthenticatedOrDie(GiottoInputMiddleware):
    """
    Put this in the input middleware stream to fail any requests that aren't
    made by authenticated users
    """
    def http(self, request):
        if not request.user:
            raise NotAuthorized('Must be Logged in for this program')
        return request

    def cmd(self, request):
        user = getattr(request, 'user', None)
        if not user:
            raise SystemExit("Must be logged in")
        return request

class NotAuthenticatedOrDie(GiottoInputMiddleware):
    """
    Put this in the input middleware stream to fail any requests that aren't
    made by authenticated users
    """
    def http(self, request):
        if request.user:
            raise NotAuthorized('Must not be Logged in for this program')
        return request

    def cmd(self, request):
        user = getattr(request, 'user', None)
        if user:
            raise SystemExit("Must not be logged in for this program")
        return request

def AuthenticatedOrRedirect(invocation):
    """
    Middleware class factory that redirects if the user is not logged in.
    Otherwise, nothing is effected.
    """
    class AuthenticatedOrRedirect(GiottoInputMiddleware):
        def http(self, request):
            if request.user:
                return request
            return Redirection(invocation)

        def cmd(self, request):
            if request.user:
                return request
            return Redirection(invocation)

    return AuthenticatedOrRedirect


def NotAuthenticatedOrRedirect(invocation):
    """
    Middleware class factory that redirects if the user is not logged in.
    Otherwise, nothing is effected.
    """
    class NotAuthenticatedOrRedirect(GiottoInputMiddleware):
        def http(self, request):
            if not request.user:
                return request
            return Redirection(invocation)

        def cmd(self, request):
            if not request.user:
                return request
            return Redirection(invocation)

    return NotAuthenticatedOrRedirect


class LogoutMiddleware(GiottoOutputMiddleware):
    """
    Logout the user. This not only deletes the session key in persistence (the
    cookie), but also nukes the session in the database so it will never
    ve valid again.
    """
    def http(self, request, response):
        if request.method == 'POST' and 'auth_session' in request.POST:
            key = request.POST['auth_session']
            session_key = get_config('auth_session').set(key, None, 1) # nuke session
        response.delete_cookie('giotto_session')
        return response


__version__ = '0.10.4'

def initialize(config):
    import giotto
    setattr(giotto, '_config', config)

def get_config(item, default=None):
    import giotto
    return getattr(giotto._config, item, default)


import traceback

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from giotto.exceptions import NoViewMethod, InvalidInput, NotAuthorized, DataNotFound, ProgramNotFound
from giotto.controllers import GiottoController
from giotto.control import Redirection
from giotto.utils import render_error_page
from webob import Request, Response
from webob.exc import (
    HTTPUnsupportedMediaType, HTTPMethodNotAllowed, HTTPTemporaryRedirect,
    HTTPNotFound, HTTPForbidden
)

http_execution_snippet = """
mock = '--model-mock' in sys.argv
from giotto.controllers.http import make_app, error_handler, serve

application = make_app(manifest, model_mock=mock)

if not config.debug:
    application = error_handler(application)

if '--run' in sys.argv:
    serve('127.0.0.1', 5000, application, use_debugger=True, use_reloader=True)"""


def serve(ip, port, application, **kwargs):
    """
    Serve a wsgi app (any wsgi app) through with either werkzeug's runserver
    or the one that comes with python.
    """
    try:
        from werkzeug.serving import run_simple
        run_simple(ip, port, application, **kwargs)
        return
    except ImportError:
        pass

    from wsgiref.simple_server import make_server
    server = make_server(ip, port, application)
    print("Serving on %s:%s" % (ip, port))
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass

def make_url(invocation, args=[], kwargs={}):
    """
    >>> make_url('some/path/program', ['arg1', 'arg2'], {'arg3': 4})
    '/some/path/program/arg1/arg2?arg3=4'
    """
    if not invocation.endswith('/'):
        invocation += '/'
    if not invocation.startswith('/'):
        invocation = '/' + invocation

    url = invocation

    for arg in args:
        url += str(arg) + "/"

    if kwargs:
        url = url[:-1]
        url += "?" + urlencode(kwargs)
    
    return url


class HTTPController(GiottoController):
    name = 'http'
    default_mimetype = 'text/html'

    def mimetype_override(self):
        accept = self.request.headers.get('Accept', '')
        has_json_in_view = self.program.view.can_render('json')
        if accept == '*/*' and self.request.is_xhr and has_json_in_view:
            # return json on ajax calls if no accept headers are present.
            # and only if the view has implemented a application/json method
            return "application/json"

        if accept != '*/*' and not accept.startswith('text/html'):
            return accept
        return None

    def get_invocation(self):
        return self.request.path

    def get_controller_name(self):
        return 'http-%s' % self.request.method.lower()

    def get_raw_data(self):
        data = {}
        if self.request.method == 'GET':
            data = self.request.GET
        elif self.request.method == 'POST':
            data = self.request.POST
        return data

    def get_concrete_response(self):
        try:
            result = self.get_data_response()
        except NoViewMethod as exc:
            return HTTPUnsupportedMediaType(
                body=render_error_page(415, exc, mimetype=self.mimetype),
                content_type="text/html"
            )
        except InvalidInput as exc:
            ## if the model raises a InvalidInput, retry the request as
            ## a GET request for the same program, and set the code to 400.
            request = make_duplicate_request(self.request)
            request.method = 'GET'
            c = HTTPController(request, self.manifest, self.model_mock, errors=exc)
            response = c.get_response()
            response.status_int = 400
            return response
        except NotAuthorized as exc:
            return HTTPForbidden(
                body=render_error_page(403, exc, mimetype=self.mimetype),
                content_type="text/html"
            )
        except (DataNotFound, ProgramNotFound) as exc:
            return HTTPNotFound(
                body=render_error_page(404, exc, mimetype=self.mimetype),
                content_type="text/html"
            )

        if type(result['body']) == Redirection:
            response = HTTPTemporaryRedirect(location=result['body'].path)
        else:
            lazy = None
            body = result['body']

            if type(body) == tuple:
                lazy = body
                body = ''

            if hasattr(body, 'read'):
                body = body.read()

            response = Response(
                status=200,
                body=body,
                content_type=result['mimetype'],
            )

            response.lazy_data = lazy

        return response

    def persist(self, persist, response):
        for key, value in persist.items():
            response.set_cookie(key, value)
        return response

    def get_primitive(self, primitive):
        if primitive == 'ALL_DATA':
            return self.get_raw_data()
        if primitive == 'LOGGED_IN_USER':
            return self.request.user
        if primitive == 'ALL_PROGRAMS':
            return self.manifest.get_all_programs()


def make_duplicate_request(request):
    """
    Since werkzeug request objects are immutable, this is needed to create an
    identical reuet object with immutable values so it can be retried after a
    POST failure.
    """
    class FakeRequest(object):
        method = 'GET'
        path = request.path
        headers = request.headers
        GET = request.GET
        POST = request.POST
        user = getattr(request, 'user', None)
        cookies = request.cookies
        is_xhr = request.is_xhr
    return FakeRequest()


def make_app(manifest, model_mock=False, cache=None):
    
    def application(environ, start_response):
        """
        WSGI app for serving giotto applications
        """
        request = Request(environ)
        controller = HTTPController(request, manifest, model_mock=model_mock)
        wsgi_response = controller.get_response()
        return wsgi_response(environ, start_response)

    return application


def error_handler(app):
    """
    WGSI middleware for catching errors and rendering the error page.
    """
    def application(environ, start_response):
        try:
            return app(environ, start_response)
        except Exception as exc:
            sio = StringIO()
            traceback.print_exc(file=sio)
            sio.seek(0)
            response =  Response(
                status=500,
                body=render_error_page(500, exc, traceback=sio.read()),
                content_type="text/html"
            )
            return response(environ, start_response)

    return application








from giotto.views import GiottoView, renders
from giotto.control import Redirection

class BasicRegisterForm(GiottoView):
    @renders('text/html')
    def html(self, result):
        return """<!DOCTYPE html>
        <html>
            <body>
                <form method="post" action="">
                    username: <input name="username" type="text"><br>
                    password: <input name="password" type="password"><br>
                    <input type="submit">
                </form>
            </body>
        </html>
        """

def LoginView(http_redirect=None):
    """
    After logging in, redirect the user to the ``http_redirect`` program.
    This is a meta program, a function that returns a program.
    ``http_redirect`` can be either a string, or a three item tuple with M objects:
    http_redirect=('/blog', [m.id], {}) | http_redirect=('/blog', [], {'name': M['name']})
    """
    class LoginView(GiottoView):
        """
        ``result`` is the session that was newly created. consult the
        ``create_session`` model for reference.
        """
        @renders('text/html')
        def html(self, result):
            ty = type(http_redirect)
            if ty == list or ty == tuple:
                assert len(http_redirect) == 3, "http_redirect must be three items"
                return Redirection(http_redirect[0], args=http_redirect[1], kwargs=http_redirect[2])
            else:
                # assume a string was passed in.
                return Redirection(http_redirect)
    return LoginView


from setuptools import setup

setup(
    name='giotto',
    version='0.10.4',
    description='Web development simplified. An MVC framework supporting Python 3.',
    long_description=open('README.rst').read(),
    author='Chris Priest',
    author_email='cp368202@ohiou.edu',
    url='https://github.com/priestc/giotto',
    packages=[
        'giotto',
        'giotto.controllers',
        'giotto.programs',
        'giotto.contrib',
        'giotto.contrib.auth',
        'giotto.contrib.static',
        'giotto.views'
    ],
    scripts=['bin/giotto'],
    include_package_data=True,
    license='LICENSE',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Programming Language :: Python",
        'Programming Language :: Python :: 2.7',
        "Programming Language :: Python :: 3",
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'webob==1.2.3',
        'six==1.2.0',
        'irc==5.0.1',
        'jinja2==2.6',
        'py-bcrypt==0.2',
        'python-mimeparse==0.1.4',
        'sqlalchemy==0.7.9',
    ],
)


class GiottoOutputMiddleware(object):
    def __init__(self, controller):
        self.controller = controller

class GiottoInputMiddleware(object):
    def __init__(self, controller):
        self.controller = controller

class RenderLazytemplate(GiottoOutputMiddleware):
    def http(self, request, response):
        engine, template, context = response.lazy_data
        if engine == 'jinja2':
            rendered = template.render(**context)
        response.text = rendered
        return response


import os
import giotto

def shell(shell=''):
    
    if shell == "bpython" or shell == 'b':
        config = giotto._config
        import bpython
        bpython.embed()
        return

    elif shell == 'ipython' or shell == 'i':
        config = giotto._config
        from IPython import embed
        embed()
        return

    else:
        import code
        imported_objects = {'config': giotto._config}
        try:  # Try activating rlcompleter, because it's handy.
            import readline
        except ImportError:
            pass
        else:
            # We don't have to wrap the following import in a 'try', because
            # we already know 'readline' was imported successfully.
            import rlcompleter
            readline.set_completer(rlcompleter.Completer(imported_objects).complete)
            readline.parse_and_bind("tab:complete")

        # We want to honor both $PYTHONSTARTUP and .pythonrc.py, so follow system
        # conventions and get $PYTHONSTARTUP first then .pythonrc.py.
        for pythonrc in (os.environ.get("PYTHONSTARTUP"),
                         os.path.expanduser('~/.pythonrc.py')):
            if pythonrc and os.path.isfile(pythonrc):
                try:
                    with open(pythonrc) as handle:
                        exec(compile(handle.read(), pythonrc, 'exec'))
                except NameError:
                    pass
        code.interact(local=imported_objects)